package com.aselimov.app.socket;

import java.io.*;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.function.Consumer;

/**
 * Listen for data over a socket
 *
 */
public class SocketListener<T> {

    private final int port;
    private final Class<T> dataClass;
    private final ObjectMapper objectMapper;
    private final Consumer<T> dataConsumer;
    private ServerSocket serverSocket;
    private boolean running = false;
    private ExecutorService executorService;

    /**
     * Creates a new StreamingSocketJsonListener.
     * 
     * @param port         The port to listen on
     * @param targetClass  The class to parse incoming JSON into
     * @param dataConsumer Consumer that processes parsed data objects
     */
    public SocketListener(int port, Class<T> dataClass, Consumer<T> dataConsumer) {
        this.port = port;
        this.dataClass = dataClass;
        this.dataConsumer = dataConsumer;
        this.objectMapper = new ObjectMapper();
    }

    /**
     * Start the SocketListener.
     * This creates a threadpool to process the socket. For each piece of data
     * received we immediately call the consumer
     * function
     */
    public void start() {
        if (running) {
            return;
        }

        System.out.println("Started listening on " + port);
        running = true;
        // Use a new Cached Thread Pool since we expect low numbers of clients
        // connecting
        executorService = Executors.newCachedThreadPool();

        executorService.execute(() -> {
            try {
                serverSocket = new ServerSocket(port);
                while (running) {
                    try {
                        Socket clientSocket = serverSocket.accept();
                        handleConnection(clientSocket);
                    } catch (IOException e) {
                        if (running) {
                            System.err.println("Error accepting connections: " + e.getMessage());
                        }
                    }
                }
            } catch (IOException e) {
                System.err.println("Error starting server because: " + e.getMessage());
            }

        });
    }

    private void handleConnection(Socket clientSocket) {
        System.out.println("Opening new client connection");
        executorService.execute(() -> {
            try {
                BufferedInputStream inputStream = new BufferedInputStream(clientSocket.getInputStream());

                JsonFactory factory = objectMapper.getFactory();
                JsonParser parser = factory.createParser(inputStream);

                // Set the parser for streaming mode
                parser.setCodec(objectMapper);

                while (running && clientSocket.isConnected() && !clientSocket.isClosed()) {
                    try {
                        T parsedData = parser.readValueAs(dataClass);

                        if (parsedData != null) {
                            // NOTE: This is going to run pretty slow for multi-client architectures because
                            // each JSON object in the stream is going to lock the data structure to push a
                            // new data point. Probably here we should use per thread buffers and then add
                            // large groups of data points at once
                            dataConsumer.accept(parsedData);
                        }
                    } catch (IOException e) {

                        if (clientSocket.isClosed() || !clientSocket.isConnected()) {
                            break;
                        } else {
                            System.err.println("Error parser JSON because " + e.getMessage());
                        }
                    }
                }
            } catch (IOException e) {
                System.err.println("Error reading from socket because " + e.getMessage());
            } finally {
                try {
                    clientSocket.close();
                    System.out.println("Closed client connection");
                } catch (IOException e) {
                    System.err.println("Error closing client socket because " + e.getMessage());
                }
            }

        });
    }

    /**
     * Stops the listener and cleans up resources.
     */
    public void stop() {
        running = false;

        System.out.println("Stopping SteadyDetect and freeing port " + port);

        if (serverSocket != null && !serverSocket.isClosed()) {
            try {
                serverSocket.close();
            } catch (IOException e) {
                System.err.println("Error closing server socket: " + e.getMessage());
            }
        }

        if (executorService != null) {
            executorService.shutdown();
        }

        System.out.println("Streaming socket listener stopped");
    }

    @Override
    protected void finalize() {
        stop();
    }

}
