package com.aselimov.app.steady_state_analysis;

import com.aselimov.app.timedata.TimeData;
import com.aselimov.app.timedata.TimeData.TimeDataPoints;
import com.aselimov.app.utilities.Statistics;

/**
 * Determine whether a time series is steady state using continuous linear
 * regression.
 *
 * This class performs a basic linear regression on chunks of the input signal.
 * If a user specified number of chunks have a slope within a specified
 * tolerance, then the signal is labeled as having reached steady state
 */
public class LinearRegression extends SteadyStateCalculator {
    private double slopeTol;
    private int numNeededMatching;
    private int windowSize;
    private int numMatching = 0;
    private int lastChecked = 0;

    /**
     * Create a LinearRegression object
     * 
     * @param slopeTol          Tolerance for the slope to be detected as 0
     * @param numNeededMatching the number of consecutive matches before we detect
     *                          the signal as steady state
     * @param windowSize        The size of the window to perform linear regression
     *                          on
     */
    public LinearRegression(double slopeTol, int numNeededMatching, int windowSize) {
        this.slopeTol = slopeTol;
        this.numNeededMatching = numNeededMatching;
        this.windowSize = windowSize;
    }

    /**
     * Determine whether a signal is steady or not.
     * This performs linear regression on the final windowSize number of data
     * points and determines whether the signal is steady state or not
     */
    @Override
    protected SignalStatus isSteady(TimeData inData) {

        // TODO: This is a race condition, in that the time between checking whether we
        // need to check steady state can yield additional data points, meaning some
        // points aren't getting checked. Probably need ot refactor this code so that
        // calling the steady checker happens on the data addition
        if ((inData.getNumPnts() - lastChecked) < windowSize) {
            return SignalStatus.NOT_CHECKED;
        }
        TimeDataPoints data = inData.threadsafeClone();
        lastChecked = data.times.size();

        // Estimate the slope of the fitted line (the intercept doesn't matter)
        int endIdx = data.times.size();
        int startIdx = endIdx - windowSize;
        double timeMean = Statistics.calcMean(data.times, startIdx, endIdx);
        double valueMean = Statistics.calcMean(data.values, startIdx, endIdx);

        double timeVariance = Statistics.calcVariance(data.times, timeMean, startIdx, endIdx);

        double covariance = Statistics.calcCovariance(data.times, data.values, timeMean, valueMean, startIdx, endIdx);

        double slope = covariance / timeVariance;

        if (Math.abs(slope) < slopeTol) {
            numMatching += 1;
        } else {
            numMatching = 0;
        }

        if (numMatching >= numNeededMatching) {
            return SignalStatus.STEADY;
        } else {
            return SignalStatus.NOT_STEADY;
        }
    }
}
